/***
*		joy.c -- joystick/controller tool for b&p
*		 1996,97 Another Option   All rights reserved.
***/

/*
=============================================================================
|   DATE   | WHAT
=============================================================================
 28 Oct 96 | created
  7 Nov 96 | by jove, I believe it works except for the gameport stuff
  8 Nov 96 | added sysex/snapshot/auto save
  2 Sep 97 | added tiny display of controllers in main window; added free sysex
           | flag (duh)
*/


#define DBUG 0

/* define this to write to sd0:BP.dump instead of con: for dprintf */
//#define FILEDUMP 666

#include <exec/types.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <intuition/intuition.h>
#include <string.h>
#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <clib/alib_protos.h>
#include <clib/alib_stdio_protos.h>
#include <stdarg.h>

#include "bars.h"
#include "edit.h"
#include "BarsMacros.h"
#include "BPLib.h"
#include "configlib.h"
#include "proto/joy_protos.h"
#include "ToolID.h"
#include "joyw.h"
#include "/icons/joy.h"
#include "configgadget2.h"
#include "snapshotgadget.h"
#include "debug.h"

struct JoyTool {
	struct Tool tool;
	struct Tool ontimetool;
	struct Event *xlist;		// list of events generated
	struct Event *xonoff;	// times for events
	struct Event *ylist;
	struct Event *yonoff;
	long xsignal;
	long ysignal;
	struct Task *task;
	UWORD flags;
	UBYTE xtype;				// type of events to generate
	UBYTE ytype;
	BYTE xcc;					// controller number
	BYTE ycc;
	UBYTE x;						// current position on axis
	UBYTE y;
	UBYTE oldx;					// previous value sent
	UBYTE oldy;
	BYTE xhi;					// range of values to send
	BYTE xlo;
	BYTE yhi;
	BYTE ylo;
	UBYTE xres;					// minimum difference between values
	UBYTE yres;
	UBYTE newx;					// potential next value to send
	UBYTE newy;
	BYTE xsignalnum;
	BYTE ysignalnum;
};

/* flag bits */
#define DEFERXB        1
#define DEFERYB        2
#define PLAYCC         4
#define AUTOSAVE       8
#define USE_SYSEX   0x10
#define SHOW_CCNUM  0x20
#define DRAGGING    0x40
#define RUNNING     0x80
#define FREE_SYSEX 0x100

/* message types to generate */
#define TYPE_OFF  0
#define TYPE_CCHG 1
#define TYPE_MTCH 2
#define TYPE_BEND 3

#define TOOL_SIZE sizeof(struct JoyTool)
extern struct IntuitionBase *IntuitionBase;
extern struct GfxBase *GfxBase;
extern struct LayersBase *LayersBase;
extern struct DOSBase *DOSBase;
extern struct ExecBase *SysBase;
extern struct Functions * functions;

static struct ToolMaster master;
static struct ToolMaster ontimemaster;

static UBYTE const versiontag[] = "$VER: OhJoy 40.11 (02.09.97)";

void AddConfigGadget(struct Window * w);

/*==========================================================================
* inittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: sets tool defaults
*=========================================================================*/

static void __regargs inittool(struct JoyTool *tool)

{
	tool->flags = AUTOSAVE | FREE_SYSEX;
	tool->xtype = tool->ytype = TYPE_CCHG;
	tool->xcc = 10;
	tool->ycc = 7;
	tool->xlo = tool->ylo = 0;
	tool->xhi = tool->yhi = 127;
	tool->x = tool->y = tool->newx = tool->newy = 64;
	tool->oldx = tool->oldy = -1;
	tool->xres = tool->yres = 5;
	tool->xlist = tool->ylist = tool->xonoff = tool->yonoff = NULL;
	tool->task = NULL;
	tool->xsignal = tool->ysignal = 0;
	tool->xsignalnum = tool->ysignalnum = 0;
	tool->tool.touched |= TOUCH_INIT;
}

/* map message types to appropriate midi status bytes */
static UBYTE status[4] = { 
	0, MIDI_CCHANGE, MIDI_MTOUCH, MIDI_PBEND
};

/************************************************************************/

static long __regargs esprintf(char *buffer, const char *command, ...)
{
	va_list arglist;

	va_start(arglist,command);
	RawDoFmt((STRPTR)command,(APTR)arglist,(void(*))("\x16\xC0\x4E\x75"),
	         (APTR)buffer);
	va_end(arglist);
	return((long)strlen(buffer));
}

struct JoyTool *selectedtool = NULL;	// active tool

#define SX_LENGTH 18

/*==========================================================================
* dosysex
*
*  INPUT: pointer to StringEvent
* RETURN: pointer to Event
*
* EFFECT: gets new tool state from sysex message
*=========================================================================*/

static __inline struct Event *dosysex(struct JoyTool *tool,
                                      struct StringEvent *event)
{
	register unsigned char *sxdata;
	struct Message *msg;
	UWORD sxflag;

	if (sxdata = checksysex(event,TOOL_SYSEX_ID,tool->tool.toolid)) {
		sxflag = tool->flags & (USE_SYSEX | FREE_SYSEX);
		memcpy((void *)&tool->flags,(void *)sxdata,SX_LENGTH);
		if (sxflag & USE_SYSEX) tool->flags |= USE_SYSEX;
		if (sxflag & FREE_SYSEX) tool->flags |= FREE_SYSEX;
		if (functions->running) tool->flags |= RUNNING;
		tool->flags &= ~DRAGGING;
		if (tool->tool.window) {
			if (msg=(struct Message *)functions->myalloc(sizeof(struct Message),0)) {
				msg->mn_Node.ln_Type = NT_MESSAGE;
				msg->mn_ReplyPort = NULL;
				msg->mn_Length = SECRET_MESSAGE;
				PutMsg(tool->tool.window->UserPort,msg);
			}
		}
		if (tool->flags & FREE_SYSEX) {
			functions->freestring(event->string);
			functions->freeevent((struct Event *)event);
			event=NULL;
		}
	}
	return((struct Event *)event);
}

/*==========================================================================
* processontime
*
*  INPUT: pointer to Event
* RETURN: pointer to Event
*
* EFFECT: processes timely events; updates tool->x/y if one of our events
*         and user is not changing it; signals edittool() if it is running 
*         and the new value exceeds the resolution.
*=========================================================================*/

static struct Event * __saveds __stdargs processontime(struct Event *event)

{
   struct JoyTool *tool = (struct JoyTool *)event->tool;

	tool = (struct JoyTool *)tool->tool.parent;
	if (tool && tool->tool.window) {
		if (
		 ((event->status==status[tool->xtype])&&(status[tool->xtype]==MIDI_CCHANGE)&&
		  (event->byte1==tool->xcc))
		 ||
		 ((status[tool->xtype]!=MIDI_CCHANGE)&&(event->status==status[tool->xtype]))
		 ) {
			if (!(tool->flags & DRAGGING) && (tool->flags & PLAYCC)) {
				if (tool->x != event->byte2) {
					if (abs(event->byte2-tool->oldx) >= tool->xres) {
						tool->newx = event->byte2;
						if (tool->task && tool->xsignal) {
							Signal(tool->task,tool->xsignal);
						}
					}
				}
			}
		}
		if (
		 ((event->status==status[tool->ytype])&&(status[tool->ytype]==MIDI_CCHANGE)&&
		  (event->byte1==tool->ycc))
		 ||
		 ((status[tool->ytype]!=MIDI_CCHANGE)&&(event->status==status[tool->ytype]))
		 ) {
			if (!(tool->flags & DRAGGING) && (tool->flags & PLAYCC)) {
				if (tool->y != event->byte2) {
					if (abs(event->byte2-tool->oldy) >= tool->yres) {
						tool->newy = event->byte2;
						if (tool->task && tool->ysignal) {
							Signal(tool->task,tool->ysignal);
						}
					}
				}
			}
		}
	}
//dprintf("ontime() free %8X\n",event);
	functions->freeevent(event);
	return(NULL);
}

/*==========================================================================
* processevent
*
*  INPUT: pointer to Event
* RETURN: pointer to Event
*
* EFFECT: processes events one at a time. frees our events if user is changing
*         joy; other events copied and sent to ontime code.
*=========================================================================*/

static struct Event * __saveds __stdargs processevent(struct Event *event)

{
   struct JoyTool *tool = (struct JoyTool *)event->tool;
	struct Event *copy;

	event->tool = event->tool->next;
	if (event->type & EVENT_PADEDIT) return(event);
	else if ((event->status == MIDI_NOTEON) || (event->status == MIDI_PTOUCH))
		return(event);
	else if (event->type & EVENT_SYSX) {
		event = dosysex(tool,(struct StringEvent *)event);
	}
	/*** EGAD, SPARKY! ***/
	else if (
	 	((event->status==status[tool->xtype])&&(status[tool->xtype]==MIDI_CCHANGE)&&
	    (event->byte1==tool->xcc))
		||
	   ((event->status==status[tool->ytype])&&(status[tool->ytype]==MIDI_CCHANGE)&&
		 (event->byte1==tool->ycc))
		|| ((status[tool->xtype]!=MIDI_CCHANGE)&&(event->status==status[tool->xtype]))
		|| ((status[tool->ytype]!=MIDI_CCHANGE)&&(event->status==status[tool->ytype]))
		) {
		if (tool->flags & DRAGGING) {
//dprintf("pe() free %8X\n",event);
			functions->freeevent(event);
			event = NULL;
		}
		else { /* don't let it go until its time */
			if (copy = functions->dupeevent(event)) {
				event->next = copy;
				copy->tool = &tool->ontimetool;
				tool->ontimetool.toolmaster = &ontimemaster;
				tool->ontimetool.parent = (struct Tool *)tool;
				tool->ontimetool.tooltype = TOOL_ONTIME | TOOL_NORMAL;
			}
		}
	}
	return(event);
}

static char *CCnames[] = {
/*0 - 4 */	"MonPres", "ModWhel", "Breath ", "MIDI 03", "Foot   ",
/*5 - 9 */	"PortTim", "DataEnt", "Volume ", "Balance", "MIDI 09",
/*10-14 */	"Pan    ", "Express", "MIDI 12", "MIDI 13", "MIDI 14",
/*15-19 */	"MIDI 15", "Genl 1 ", "Genl 2 ", "Genl 3 ", "Genl 4 ",
/*20-24 */	"MIDI 20", "MIDI 21", "MIDI 22", "MIDI 23", "MIDI 24",
/*25-29 */	"MIDI 25", "MIDI 26", "MIDI 27", "MIDI 28", "MIDI 29",
/*30-34 */	"MIDI 30", "MIDI 31", "MIDI 32", "MIDI 33", "MIDI 34",
/*35-39 */	"MIDI 35", "MIDI 36", "MIDI 37", "MIDI 38", "MIDI 39",
/*40-44 */	"MIDI 40", "MIDI 41", "MIDI 42", "MIDI 43", "MIDI 44",
/*45-49 */	"MIDI 45", "MIDI 46", "MIDI 47", "MIDI 48", "MIDI 49",
/*50-54 */	"MIDI 50", "MIDI 51", "MIDI 52", "MIDI 53", "MIDI 54",
/*55-59 */	"MIDI 55", "MIDI 56", "MIDI 57", "MIDI 58", "MIDI 59",
/*60-64 */	"MIDI 60", "MIDI 61", "MIDI 62", "MIDI 63", "Sustain",
/*65-69 */	"PortSwt", "Sosten.", "SoftPed", "MIDI 68", "FrezPdl",
/*70-74 */	"MIDI 70", "MIDI 71", "MIDI 72", "MIDI 73", "MIDI 74",
/*75-79 */	"LegtoSw", "MIDI 76", "MIDI 77", "MIDI 78", "MIDI 79",
/*80-84 */	"Genl 5 ", "Genl 6 ", "Genl 7 ", "Genl 8 ", "MIDI 84",
/*85-89 */	"MIDI 85", "MIDI 86", "MIDI 87", "MIDI 88", "MIDI 89",
/*90-94 */	"MIDI 90", "FXDepth", "TremDep", "ChorDep", "CelDep ",
/*95-99 */	"PhsrDep", "DataInc", "DataDec", "MIDI 98", "MIDI 99",
/*100-104*/	"MIDI100", "MIDI101", "MIDI102", "MIDI103", "MIDI104",
/*105-109*/	"MIDI105", "MIDI106", "MIDI107", "MIDI108", "MIDI109",
/*110-114*/	"MIDI110", "MIDI111", "MIDI112", "MIDI113", "MIDI114",
/*115-119*/	"MIDI115", "MIDI116", "MIDI117", "MIDI118", "MIDI119",
/*120-124*/	"MIDI120", "RsetCtl"/*, "LoclCtl", "AllOff ", "OmniOff",*/
/*125-127*/	/*"Omni On", "Mono On", "Poly On"*/
};

/*==========================================================================
* hiloCCName
*
*  INPUT: pointers to window and gadget, gadget value
* RETURN: gadget value
*
* EFFECT: display a pot value mapped to controller name
*=========================================================================*/

static long __saveds __stdargs hiloCCName(struct Window *w, struct Gadget *g, long n)
{
	char text[8];

	SetRPort(w->RPort, g->LeftEdge+g->Width+XOffset, g->TopEdge+YOffset,
	         (g->Flags & GADGDISABLED) ? BP_DGREY : BP_WHITE, 0, JAM2);
	if (selectedtool->flags & SHOW_CCNUM) {
		esprintf(text,"%3ld",n);
		strcat(&text[3],"    ");
		Text(w->RPort,text,7);
	}
	else {
		Text(w->RPort,CCnames[n],7);
	}
	return(n);
}


/*==========================================================================
* hiloType
*
*  INPUT: pointers to window and gadget, gadget value
* RETURN: gadget value
*
* EFFECT: display a pot value mapped to event type
*=========================================================================*/

static long __saveds __stdargs hiloType(w,g,h)

struct Window *w;
struct Gadget *g;
long h;

{
	static char *eventtypes[] = {
		"Off   ",
		"CtlChg",  /* type 1 */
		"MTouch",  /* type 2 */
		"PBend "   /* type 3 */
	};

	SetRPort(w->RPort,g->LeftEdge+g->Width+XOffset-1,g->TopEdge+YOffset,
            HILO_APEN,HILO_BPEN,HILO_DRAWMODE);
	Text(w->RPort,eventtypes[h],6);
	return(h);
}

/* flags for which axis we're sending from */
#define SEND_X 1
#define SEND_Y 2

/*==========================================================================
* sendcontrol
*
*  INPUT: pointer to JoyTool, cc#, cc value
* RETURN: none
*
* EFFECT: queues and links controller messages 
*=========================================================================*/

static void __regargs sendcontrol(struct JoyTool *tool, UBYTE flag,
                                  UBYTE status, UBYTE b1, UBYTE b2)
{
	struct Event *event;

	event = functions->allocevent();
	if (event) { // send the event
		event->tool = tool->tool.next;
		event->type = EVENT_VOICE | EVENT_PLAYONLY; // don't record; we be splicing
		event->time = functions->timenow;
		event->status = status;
		if (status == MIDI_CCHANGE) {
			event->byte1 = b1;
			event->byte2 = b2;
		}
		else if (status == MIDI_MTOUCH) { // this sends only 7-bit messages
			event->byte1 = b2;
			event->byte2 = 0;
		}
		else { // better be pitch bend
			event->byte1 = 0;
			event->byte2 = b2;
		}
		functions->qevent(event);
	}
	if (tool->flags & RUNNING) { // save one to splice into clip
		event = functions->allocevent();
		if (event) {
			event->time = functions->timenow;
			event->status = status;
			if (status == MIDI_CCHANGE) {
				event->byte1 = b1;
				event->byte2 = b2;
			}
			else if (status == MIDI_MTOUCH) {
				event->byte1 = b2;
				event->byte2 = 0;
			}
			else {
				event->byte1 = 0;
				event->byte2 = b2;
			}
			if (flag == SEND_X) {
				event->next = tool->xlist;
				tool->xlist = event;
			}
			else if (flag == SEND_Y) {
				event->next = tool->ylist;
				tool->ylist = event;
			}
		}
	}
}

/*==========================================================================
* dosnapshot
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: makes a sysex message and posts it
*=========================================================================*/

static void dosnapshot(struct JoyTool *tool)

{
	UBYTE sx[SX_LENGTH];
	struct StringEvent *event;

	memcpy((void *)sx,(void *)&tool->flags,SX_LENGTH); 
	sx[1] = (UBYTE)(tool->flags & ~(DRAGGING|RUNNING)); /* mayhem if restored */
	if (event = BuildSnapshot(&tool->tool, 
                             functions->running ? functions->timenow : -1,
		                       sx, SX_LENGTH)) {
		PostSnapshot(&tool->tool,event);
		if (tool->tool.track->edit->window)
			functions->updateobject(&functions->inittempo,0); /* redisplay */
	}
}

/* coordinates for joystick values display */
#define XVAL_X 196
#define XVAL_Y  86
#define YVAL_X XVAL_X 
#define YVAL_Y  96

/* for reading & typing ease */
#define YVal(y) (selectedtool->yhi - y)
#define XVal(x) (x + selectedtool->xlo)

/*==========================================================================
* hiloJoy
*
*  INPUT: the usual
* RETURN: 
*
* EFFECT: updates tool x/y value if non-redundant; calls sendcontrol() to 
*         send cc; updates gadget value display
*=========================================================================*/

static long __saveds __stdargs hiloJoy(struct Window *w,struct Gadget *g,long x,long y)
{
	char text[6];

/*
dprintf("xin = %3ld yin = %3ld\n",x,y);
dprintf("oldx=%3d newx=%3d oldy=%3d newy=%3d\n",
     selectedtool->oldx,selectedtool->newx,selectedtool->oldy,selectedtool->newy);
*/
	if ((abs(selectedtool->oldx - XVal(x)) >= selectedtool->xres) ||
		((XVal(x) == selectedtool->xlo || XVal(x) == selectedtool->xhi) && 
		 (XVal(x) != selectedtool->oldx)))
	{
		selectedtool->x = XVal(x);
		/* send contoller if this is us & it's not redundant */
		if (selectedtool && (w==selectedtool->tool.window) && selectedtool->xtype) {
			if (!(selectedtool->flags & (DEFERXB | PLAYCC))) {
				sendcontrol(selectedtool,SEND_X,status[selectedtool->xtype],
				            selectedtool->xcc,selectedtool->x);
				selectedtool->oldx = selectedtool->newx = selectedtool->x;
			}
			else selectedtool->newx = XVal(x);
		}
	}
	else selectedtool->newx = XVal(x);
	esprintf(text,"x=%3ld",(selectedtool->newx <= selectedtool->xhi) ?
	                        selectedtool->newx : selectedtool->xhi);
	SetRPort(w->RPort,XVAL_X,XVAL_Y,selectedtool->xtype ? HILO_APEN : HILO_DPEN,
	         HILO_BPEN,HILO_DRAWMODE);
	Text(w->RPort,text,5);

	if ((abs((selectedtool->yhi-selectedtool->oldy)-(y+selectedtool->ylo)) >= 
	          selectedtool->yres) ||
		 ((YVal(y) == selectedtool->ylo || YVal(y) == selectedtool->yhi) &&
		  (YVal(y) != selectedtool->oldy)))
	{
		selectedtool->y = YVal(y);
		if (selectedtool && (w==selectedtool->tool.window) && selectedtool->ytype) {
			if (!(selectedtool->flags & (DEFERYB | PLAYCC))) {
				sendcontrol(selectedtool,SEND_Y,status[selectedtool->ytype],
				            selectedtool->ycc,selectedtool->y);
				selectedtool->oldy = selectedtool->newy = selectedtool->y;
			}
			else selectedtool->newy = YVal(y);
		}
	}
	else selectedtool->newy = YVal(y);
	esprintf(text,"y=%3ld",(selectedtool->newy <= selectedtool->yhi) ?
	                        selectedtool->newy : selectedtool->yhi);
	SetRPort(w->RPort,YVAL_X,YVAL_Y,selectedtool->ytype ? HILO_APEN : HILO_DPEN,
	         HILO_BPEN,HILO_DRAWMODE);
	Text(w->RPort,text,5);
	return(0);
}

/*==========================================================================
* onoff
*
*  INPUT: 
* RETURN: 
*
* EFFECT: stashes a list of events with the times the joy gadget got dragged
*         (byte2==1) and released (byte2==0) to sort out times for splicing
*         the stuff into the track's clip later.
*=========================================================================*/

static void __regargs onoff(struct JoyTool *tool, struct Event **onoff,
                            UBYTE status, UBYTE cc, UBYTE on)
{
	struct Event *event;

	if ((tool->flags & RUNNING) && status) {
		if (event = functions->allocevent()) {
			event->time = functions->timenow;
			event->status = status;
			event->byte1 = cc;
			event->byte2 = on;
			event->next = *onoff;
			*onoff = event;
		}
	}
}

#define tw tool->tool.window

// show ccs in small font on main area of editwindow 
void __inline showcontrollerassignment(struct JoyTool *tool)
{
	char text[11];

	/* clear space with fixed font; tiny font is proportional (damnit) */
	memset(text,0x20,6);
	SetRPort(tw->RPort,11,88,0,0,JAM2);
	Text(tw->RPort,text,6);
	SetRPort(tw->RPort,11,97,0,0,JAM2);
	Text(tw->RPort,text,6);
	if (tool->flags & SHOW_CCNUM) esprintf(text,"X: %3ld    ",tool->xcc);
	else esprintf(text,"X: %s",CCnames[tool->xcc]);
	functions->usesmallfont(tw->RPort);
	SetRPort(tw->RPort,11,88,BP_LGREY,0,JAM2);
	Text(tw->RPort,text,10);
	if (tool->flags & SHOW_CCNUM) esprintf(text,"Y: %3ld    ",tool->ycc);
	else esprintf(text,"Y: %s",CCnames[tool->ycc]);
	SetRPort(tw->RPort,11,97,BP_LGREY,0,JAM2);
	Text(tw->RPort,text,10);
	functions->usetopazfont(tw->RPort);
}

/* refresh bits */
#define REFRESH_BOOL    1
#define REFRESH_XBOOL   2
#define REFRESH_YBOOL   4
#define REFRESH_JOY     8
#define REFRESH_XPOT   16
#define REFRESH_YPOT   32
#define REFRESH_CONFIG 64

/*==========================================================================
* edittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: ui handler
*=========================================================================*/

static void __saveds __stdargs edittool(struct JoyTool *tool)

{
	struct IntuiMessage *message;
	struct Gadget *gadget;
	long class; 
	long code;
	struct Menu *menu;
	struct MenuItem *menuitem;
	struct NewWindow *nw;
	long signal;
	UBYTE refresh; // also used to hold InToolPad() return in various places!

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);
	refresh = InToolPad((struct Tool *)tool);
	/* get menus and flags in sync before duping menu */
	if (tool->flags & SHOW_CCNUM) MenuItem1.Flags |= CHECKED;
	else MenuItem1.Flags &= ~CHECKED;
	if (tool->flags & USE_SYSEX) MenuItem2.Flags |= CHECKED;
	else MenuItem2.Flags &= ~CHECKED;
	if (tool->flags & PLAYCC) MenuItem3.Flags |= CHECKED;
	else MenuItem3.Flags &= ~CHECKED;
	if (tool->flags & AUTOSAVE) MenuItem4.Flags |= CHECKED;
	else MenuItem4.Flags &= ~CHECKED;
	if (refresh) MenuItem5.Flags &= ~MENUENABLED;
	else MenuItem5.Flags |= MENUENABLED;
	nw = OpenEmbossedWindowSlop(&NewWindowStructure1,&tool->tool,0,WTITLE,
	                            refresh ? 20 : 40);
	if (!nw) return;
	AddConfigGadget(tw);
	if (!refresh) AddSnapshotGadget(tw);
	menu = functions->DupeMenu(&MenuList1);
	if (menu) SetMenuStrip(tw,menu);
	else SetWindowTitles(tw,(UBYTE *)-1,MenuList1.MenuName);
	tool->task = FindTask(NULL);
	tool->xsignalnum = AllocSignal(-1);
	tool->xsignal = (1L << tool->xsignalnum);
	tool->ysignalnum = AllocSignal(-1);
	tool->ysignal = (1L << tool->ysignalnum);
	for (refresh=DOIT ; refresh <= GCOUNT ; refresh++) {
		functions->EmbossOn(tw,refresh,1);
	}
	functions->EmbossedPropOn(tw,JOY,hiloJoy,128,128);
	functions->FatEmbossedPropOn(tw,XTYPE,XTYPE+GDOWN,XTYPE+GUP,hiloType,4,1);
	functions->FatEmbossedPropOn(tw,XCC,XCC+GDOWN,XCC+GUP,hiloCCName,121,1);
	functions->FatEmbossedPropOn(tw,XHI,XHI+GDOWN,XHI+GUP,hilo,128,1);
	functions->FatEmbossedPropOn(tw,XLO,XLO+GDOWN,XLO+GUP,hilo,128,1);
	functions->FatEmbossedPropOn(tw,XRES,XRES+GDOWN,XRES+GUP,hilo,21,1);
	functions->FatEmbossedPropOn(tw,YTYPE,YTYPE+GDOWN,YTYPE+GUP,hiloType,4,1);
	functions->FatEmbossedPropOn(tw,YCC,YCC+GDOWN,YCC+GUP,hiloCCName,121,1);
	functions->FatEmbossedPropOn(tw,YHI,YHI+GDOWN,YHI+GUP,hilo,128,1);
	functions->FatEmbossedPropOn(tw,YLO,YLO+GDOWN,YLO+GUP,hilo,128,1);
	functions->FatEmbossedPropOn(tw,YRES,YRES+GDOWN,YRES+GUP,hilo,21,1);
	refresh = REFRESH_BOOL | REFRESH_XBOOL | REFRESH_YBOOL | REFRESH_JOY;
	if (WindowIsTall(tw)) refresh |= REFRESH_XPOT | REFRESH_YPOT | REFRESH_CONFIG;

	/*** ui handling ***/
	for ( ; ; ) {
		if (refresh) {
			showcontrollerassignment(tool);
			if (refresh & REFRESH_BOOL) {
				functions->SelectEmbossed(tw,PLAYBACK,tool->flags & PLAYCC);
				DrawEmbossedGList(tw,functions->GetGadget(tw,DOIT),5);
			}
			if (refresh & REFRESH_XBOOL) {
				functions->SelectEmbossed(tw,DEFERX,tool->flags & DEFERXB);
				for (code=DEFERX ; code < DEFERX+5 ; code++) {
					functions->EnableEmbossed(tw,code,tool->xtype);
				}
				DrawEmbossedGList(tw,functions->GetGadget(tw,DEFERX),5);
			}
			if (refresh & REFRESH_YBOOL) {
				functions->SelectEmbossed(tw,DEFERY,tool->flags & DEFERYB);
				for (code=DEFERY ; code < DEFERY+5 ; code++) {
					functions->EnableEmbossed(tw,code,tool->ytype);
				}
				DrawEmbossedGList(tw,functions->GetGadget(tw,DEFERY),5);
			}
			if ((refresh & REFRESH_XPOT) && WindowIsTall(tw)) {
				functions->ModifyEmbossedProp(tw,XTYPE,tool->xtype,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,XTYPE);
				selectedtool = tool;
				EnableFatProp(tw,XCC,XCC+GDOWN,XCC+GUP,tool->xtype==TYPE_CCHG);
				functions->ModifyEmbossedProp(tw,XCC,tool->xcc,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,XCC);
				selectedtool = NULL;
				if (tool->xhi < tool->xlo) tool->xhi = tool->xlo+1;
				if (tool->xhi > 127) {
					tool->xhi = 127;
					tool->xlo = 126;
				}
				if (tool->xlo >= tool->xhi) tool->xlo = tool->xhi-1;
				if (tool->xlo < 0) {
					tool->xlo = 0;
					tool->xhi = 1;
				}
				functions->ModifyEmbossedProp(tw,XHI,tool->xhi,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,XHI);
				functions->ModifyEmbossedProp(tw,XLO,tool->xlo,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,XLO);
				functions->ModifyEmbossedProp(tw,XRES,tool->xres,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,XRES);
			}
			if ((refresh & REFRESH_YPOT) && WindowIsTall(tw)) {
				functions->ModifyEmbossedProp(tw,YTYPE,tool->ytype,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,YTYPE);
				selectedtool = tool;
				EnableFatProp(tw,YCC,YCC+GDOWN,YCC+GUP,tool->ytype==TYPE_CCHG);
				functions->ModifyEmbossedProp(tw,YCC,tool->ycc,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,YCC);
				selectedtool = NULL;
				if (tool->yhi < tool->ylo) tool->yhi = tool->ylo+1;
				if (tool->yhi < 0) {
					tool->yhi = 127;
					tool->ylo = 126;
				}
				if (tool->ylo >= tool->yhi) tool->ylo = tool->yhi-1;
				if (tool->ylo < 0) {
					tool->ylo = 0;
					tool->yhi = 1;
				}
				functions->ModifyEmbossedProp(tw,YHI,tool->yhi,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,YHI);
				functions->ModifyEmbossedProp(tw,YLO,tool->ylo,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,YLO);
				functions->ModifyEmbossedProp(tw,YRES,tool->yres,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,YRES);
			}
			if (refresh & REFRESH_JOY) {
				selectedtool = tool;
#pragma msg 100 ignore push
				functions->ModifyEmbossedProp(tw,JOY,
				                       tool->newx - tool->xlo, tool->yhi - tool->newy,
				                       tool->xhi-tool->xlo+1, tool->yhi-tool->ylo+1,
				                       0,0);
/* changing the move fucks up the gadget!
		                                    IDivU(tool->xhi-tool->xlo,10),
		                                    IDivU(tool->yhi-tool->ylo,10));
*/
#pragma msg 100 ignore pop
				functions->DrawEmbossedProp(tw,JOY);
				selectedtool = NULL;
/*
dprintf("x=%3d y=%3d xlo=%3d xhi=%3d oldx=%3d newx=%3d ylo=%3d yhi=%3d oldy=%3d newy=%3d\n",
         tool->x,tool->y,tool->xlo,tool->xhi,tool->oldx,tool->newx,tool->ylo,tool->yhi,tool->oldy,tool->newy);
*/
			}
			if ((refresh & REFRESH_CONFIG) && WindowIsTall(tw)) {
				DrawBevelBox(tw->RPort,X_X1,X_Y1,X_X2,X_Y2,-1,-1,1);
				DrawBevelBox(tw->RPort,Y_X1,Y_Y1,Y_X2,Y_Y2,-1,-1,1);
				PrintIText(tw->RPort,&IntuiTextList1,0,0);
				/* redraw all the pot clickers */
				for (refresh=YTYPE ; refresh < XRES+1 ; refresh++ ) {
					functions->DrawEmbossed(tw,refresh+GDOWN);
					functions->DrawEmbossed(tw,refresh+GUP);
				}
			}
			refresh = 0;
		}
		message = (struct IntuiMessage *)GetMsg(tw->UserPort);
		if (!message) {
			signal = Wait(-1);
			message = (struct IntuiMessage *)GetMsg(tw->UserPort);
		}
		if (message) { /* trick setup that probably won't work, but does! */
			if (message->ExecMessage.mn_Length == SECRET_MESSAGE) {
				/* this means we read a sysex event */
				refresh = 0xFF;
				functions->myfree((char *)message,sizeof(struct Message));
				continue;
			}
		}
		else if (!message) {
			if (signal & tool->xsignal) {
//dprintf("xsignal x=%3d newx=%3d\n",tool->x,tool->newx);
				if (tool->newx != tool->x) {
					tool->x = tool->newx;
					refresh = REFRESH_JOY;
				}
				continue;
			}
			if (signal & tool->ysignal) {
//dprintf("ysignal y=%3d newy=%3d\n",tool->y,tool->newy);
				if (tool->newy != tool->y) {
					tool->y = tool->newy;
					refresh = REFRESH_JOY;
				}
				continue;
			}
			continue;
		}
		class = message->Class;
		code = message->Code;
		gadget = (struct Gadget *)message->IAddress;
		ReplyMsg((struct Message *)message);
		class = functions->SystemGadgets(tw,class,gadget,code);

		if (class == CLOSEWINDOW) {
			tool->task = NULL;
			selectedtool = NULL;
			FreeSignal(tool->xsignalnum);
			FreeSignal(tool->ysignalnum);
			tool->xsignal = tool->ysignal = 0;
			break;
		}
		else if (class == NEWSIZE) {
			/* make room for config gadget & maybe snap gadget again */
			refresh = InToolPad((struct Tool *)tool);
			SizeTitleBar(tw,refresh ? -20 : -40);
			RefreshGList(functions->GetGadget(tw,CONFIG_GADGET_ID),tw,NULL,1);
			if (!refresh) RefreshGList(functions->GetGadget(tw,SNAPSHOT_GADGET_ID),tw,NULL,1);
			if (WindowIsTall(tw)) {
				refresh = REFRESH_XPOT | REFRESH_YPOT | REFRESH_CONFIG;
			}
		}

		else if (class == MENUPICK) {
			while (code != MENUNULL) {
				menuitem = ItemAddress(menu,code);
				switch (ITEMNUM(code)) {
					case 0: /* show cc nums */
						tool->flags ^= SHOW_CCNUM;
						refresh = REFRESH_XPOT | REFRESH_YPOT;
						break;
					case 1: /* use sysex */
						tool->flags ^= USE_SYSEX;
						break;
					case 2: /* play ccs */
						tool->flags ^= PLAYCC;
						refresh = REFRESH_BOOL;
						break;
					case 3: /* save ccs without confirming */
						tool->flags ^= AUTOSAVE;
						break;
					case 4: /* snapshot */
						dosnapshot(tool);
						break;
				}
				code = menuitem->NextSelect;
			}
		}

		else if (class == GADGETUP) {
			switch (gadget->GadgetID) {
				case DEFERX :
					tool->flags ^= DEFERXB;
					refresh = REFRESH_XBOOL;
					break;
				case DEFERY :
					tool->flags ^= DEFERYB;
					refresh = REFRESH_YBOOL;
					break;
				case PLAYBACK :
					tool->flags ^= PLAYCC;
					if (tool->flags & PLAYCC) {
						tw->MenuStrip->FirstItem->NextItem->NextItem->Flags |= CHECKED;
					}
					else {
						tw->MenuStrip->FirstItem->NextItem->NextItem->Flags &= ~CHECKED;
					}
					refresh = REFRESH_BOOL;
					break;
				case CONFIG_GADGET_ID :
					ZoomWindow(tw);
					break;
				case SNAPSHOT_GADGET_ID :
					dosnapshot(tool);
					break;
			}
		}

		else if (class == GADGETDOWN) {
			switch (gadget->GadgetID) {
				case JOY :
					tool->flags |= DRAGGING;
					if (!(tool->flags & PLAYCC)) {
						onoff(tool,&tool->xonoff,status[tool->xtype],tool->xcc,1);
						onoff(tool,&tool->yonoff,status[tool->ytype],tool->ycc,1);
					}
					selectedtool = tool;
					functions->DragEmbossedProp(tw,JOY);
					selectedtool = NULL;
					if (!(tool->flags & PLAYCC)) {
						onoff(tool,&tool->xonoff,status[tool->xtype],tool->xcc,0);
						onoff(tool,&tool->yonoff,status[tool->ytype],tool->ycc,0);
					}
					tool->flags &= ~DRAGGING;
					refresh = REFRESH_JOY;
					break;

				case MIN :
					tool->newx = tool->xlo;
					tool->newy = tool->ylo;
					refresh = REFRESH_JOY;
					break;
				case MID :
					tool->newx = ((tool->xhi-tool->xlo+1)>>1)+tool->xlo;
					tool->newy = ((tool->yhi-tool->ylo+1)>>1)+tool->ylo;
					refresh = REFRESH_JOY;
					break;
				case MAX :
					tool->newx = tool->xhi;
					tool->newy = tool->yhi;
					refresh = REFRESH_JOY;
					break;
				case DOIT :
					if (tool->xtype) {
						if (tool->newx > tool->xhi) tool->newx = tool->xhi;
						onoff(tool,&tool->xonoff,status[tool->xtype],tool->xcc,1);
						sendcontrol(tool,SEND_X,status[tool->xtype],tool->xcc,tool->newx);
						onoff(tool,&tool->xonoff,status[tool->xtype],tool->xcc,0);
						tool->oldx = tool->x = tool->newx;
						refresh = REFRESH_BOOL | REFRESH_JOY;
					}
					if (tool->ytype) {
						if (tool->newy > tool->yhi) tool->newy = tool->yhi;
						onoff(tool,&tool->yonoff,status[tool->ytype],tool->ycc,1);
						sendcontrol(tool,SEND_Y,status[tool->ytype],tool->ycc,tool->newy);
						onoff(tool,&tool->yonoff,status[tool->ytype],tool->ycc,0);
						tool->oldy = tool->y = tool->newy;
						refresh = REFRESH_BOOL | REFRESH_JOY;
					}
					break;

				case SENDX :
					if (tool->newx > tool->xhi) tool->newx = tool->xhi;
					onoff(tool,&tool->xonoff,status[tool->xtype],tool->xcc,1);
					sendcontrol(tool,SEND_X,status[tool->xtype],tool->xcc,tool->newx);
					onoff(tool,&tool->xonoff,status[tool->xtype],tool->xcc,0);
					tool->oldx = tool->x = tool->newx;
					refresh = REFRESH_XBOOL | REFRESH_JOY;
					break;
				case MINX :
					tool->newx = tool->xlo;
					refresh = REFRESH_JOY;
					break;
				case MIDX :
					tool->newx = ((tool->xhi-tool->xlo+1)/2)+tool->xlo;
					refresh = REFRESH_JOY;
					break;
				case MAXX :
					tool->newx = tool->xhi;
					refresh = REFRESH_JOY;
					break;

				case SENDY :
					if (tool->newy > tool->yhi) tool->newy = tool->yhi;
					onoff(tool,&tool->yonoff,status[tool->ytype],tool->ycc,1);
					sendcontrol(tool,SEND_Y,status[tool->ytype],tool->ycc,tool->newy);
					onoff(tool,&tool->yonoff,status[tool->ytype],tool->ycc,0);
					tool->oldy = tool->y = tool->newy;
					refresh = REFRESH_YBOOL | REFRESH_JOY;
					break;
				case MINY :
					tool->newy = tool->ylo;
					refresh = REFRESH_JOY;
					break;
				case MIDY :
					tool->newy = ((tool->yhi-tool->ylo+1)/2)+tool->ylo;
					refresh = REFRESH_JOY;
					break;
				case MAXY :
					tool->newy = tool->yhi;
					refresh = REFRESH_JOY;
					break;

				case XTYPE :
					tool->xtype = functions->DragEmbossedProp(tw,XTYPE);
					refresh = REFRESH_XPOT | REFRESH_XBOOL | REFRESH_JOY;
					break;
				case XTYPE+GDOWN :
					tool->xtype = functions->ShiftEmbossedProp(tw,XTYPE,-1,0);
					refresh = REFRESH_XPOT | REFRESH_XBOOL | REFRESH_JOY;
					break;
				case XTYPE+GUP :
					tool->xtype = functions->ShiftEmbossedProp(tw,XTYPE,1,0);
					refresh = REFRESH_XPOT | REFRESH_XBOOL | REFRESH_JOY;
					break;
				case XCC :
					selectedtool = tool;
					tool->xcc = functions->DragEmbossedProp(tw,XCC);
					refresh = REFRESH_XPOT;
					break;
				case XCC+GDOWN :
					selectedtool = tool;
					tool->xcc = functions->ShiftEmbossedProp(tw,XCC,-1,0);
					refresh = REFRESH_XPOT;
					break;
				case XCC+GUP :
					selectedtool = tool;
					tool->xcc = functions->ShiftEmbossedProp(tw,XCC,1,0);
					refresh = REFRESH_XPOT;
					break;
				case XHI :
					tool->xhi = functions->DragEmbossedProp(tw,XHI);
					refresh = REFRESH_XPOT | REFRESH_JOY;
					break;
				case XHI+GDOWN :
					tool->xhi = functions->ShiftEmbossedProp(tw,XHI,-1,0);
					refresh = REFRESH_XPOT | REFRESH_JOY;
					break;
				case XHI+GUP :
					tool->xhi = functions->ShiftEmbossedProp(tw,XHI,1,0);
					refresh = REFRESH_XPOT | REFRESH_JOY;
					break;
				case XLO :
					tool->xlo = functions->DragEmbossedProp(tw,XLO);
					refresh = REFRESH_XPOT | REFRESH_JOY;
					break;
				case XLO+GDOWN :
					tool->xlo = functions->ShiftEmbossedProp(tw,XLO,-1,0);
					refresh = REFRESH_XPOT | REFRESH_JOY;
					break;
				case XLO+GUP :
					tool->xlo = functions->ShiftEmbossedProp(tw,XLO,1,0);
					refresh = REFRESH_XPOT | REFRESH_JOY;
					break;
				case XRES :
					tool->xres = functions->DragEmbossedProp(tw,XRES);
					refresh = REFRESH_XPOT;
					break;
				case XRES+GDOWN :
					tool->xres = functions->ShiftEmbossedProp(tw,XRES,-1,0);
					refresh = REFRESH_XPOT;
					break;
				case XRES+GUP :
					tool->xres = functions->ShiftEmbossedProp(tw,XRES,1,0);
					refresh = REFRESH_XPOT;
					break;


				case YTYPE :
					tool->ytype = functions->DragEmbossedProp(tw,YTYPE);
					refresh = REFRESH_YPOT | REFRESH_YBOOL | REFRESH_JOY;
					break;
				case YTYPE+GDOWN :
					tool->ytype = functions->ShiftEmbossedProp(tw,YTYPE,-1,0);
					refresh = REFRESH_YPOT | REFRESH_YBOOL | REFRESH_JOY;
					break;
				case YTYPE+GUP :
					tool->ytype = functions->ShiftEmbossedProp(tw,YTYPE,1,0);
					refresh = REFRESH_YPOT | REFRESH_YBOOL | REFRESH_JOY;
					break;
				case YCC :
					selectedtool = tool;
					tool->ycc = functions->DragEmbossedProp(tw,YCC);
					refresh = REFRESH_YPOT;
					break;
				case YCC+GDOWN :
					selectedtool = tool;
					tool->ycc = functions->ShiftEmbossedProp(tw,YCC,-1,0);
					refresh = REFRESH_YPOT;
					break;
				case YCC+GUP :
					selectedtool = tool;
					tool->ycc = functions->ShiftEmbossedProp(tw,YCC,1,0);
					refresh = REFRESH_YPOT;
					break;
				case YHI :
					tool->yhi = functions->DragEmbossedProp(tw,YHI);
					refresh = REFRESH_YPOT | REFRESH_JOY;
					break;
				case YHI+GDOWN :
					tool->yhi = functions->ShiftEmbossedProp(tw,YHI,-1,0);
					refresh = REFRESH_YPOT | REFRESH_JOY;
					break;
				case YHI+GUP :
					tool->yhi = functions->ShiftEmbossedProp(tw,YHI,1,0);
					refresh = REFRESH_YPOT | REFRESH_JOY;
					break;
				case YLO :
					tool->ylo = functions->DragEmbossedProp(tw,YLO);
					refresh = REFRESH_YPOT | REFRESH_JOY;
					break;
				case YLO+GDOWN :
					tool->ylo = functions->ShiftEmbossedProp(tw,YLO,-1,0);
					refresh = REFRESH_YPOT | REFRESH_JOY;
					break;
				case YLO+GUP :
					tool->ylo = functions->ShiftEmbossedProp(tw,YLO,1,0);
					refresh = REFRESH_YPOT | REFRESH_JOY;
					break;
				case YRES :
					tool->yres = functions->DragEmbossedProp(tw,YRES);
					refresh = REFRESH_YPOT;
					break;
				case YRES+GDOWN :
					tool->yres = functions->ShiftEmbossedProp(tw,YRES,-1,0);
					refresh = REFRESH_YPOT;
					break;
				case YRES+GUP :
					tool->yres = functions->ShiftEmbossedProp(tw,YRES,1,0);
					refresh = REFRESH_YPOT;
					break;
			}
		}
	}

	EmbossOffGList(tw,functions->GetGadget(tw,MIN),GCOUNT);
	functions->EmbossedPropOff(tw,JOY);
	for (refresh=YTYPE ; refresh < XRES+1 ; refresh++ ) {
		functions->FatEmbossedPropOff(tw,refresh,refresh+GDOWN,refresh+GUP);
	}
	if (menu) functions->DeleteMenu(menu);
	NewCloseEmbossedWindow(&tool->tool,nw);
}

/*==========================================================================
* loadtool
*
*  INPUT: file id, size of chunk to read
* RETURN: pointer to allocated tool, or NULL
*
* EFFECT: allocates and initializes a tool with data read from disk. Resets
*         pointers and signals to avoid catastrophe.
*=========================================================================*/

static struct Tool * __saveds __stdargs loadtool(long file, long size)

{
	struct JoyTool *tool;

	if (tool = (struct JoyTool *)functions->myalloc(TOOL_SIZE,0)) {
		functions->fastread(file,(char *)tool,TOOL_SIZE);
		tool->tool.window = NULL;
		tool->xlist = tool->ylist = tool->xonoff = tool->yonoff = NULL;
		tool->task = NULL;
		tool->xsignal = tool->ysignal = 0;
		tool->xsignalnum = tool->ysignalnum = 0;
//		tool->tool.touched &= ~TOUCH_REFRESHOK;
		size -= TOOL_SIZE;
	}
	if (size) { /* leftovers */
		functions->fastseek(file,size,0);
	}
	return((struct Tool *)tool);
}

/*==========================================================================
* splicelist
*
*  INPUT: 
* RETURN: 
*
* EFFECT: puts the list into the clip
*=========================================================================*/

static void __regargs splicelist(struct Clip *clip, struct Event *list,
                                 struct Event *onoff, UBYTE status, UBYTE cc,
											UBYTE replace)

{
	long begin, end;
	register struct Event *scan;
	register struct Event *time;

	onoff = functions->sorteventlist(onoff);
/*
if (onoff) {
dprintf("splice() list=%8X onofflist=\n",list);
for (scan=onoff ; scan ; scan=scan->next) {
  dprintf("  event=%8X time=%8ld status=%2X b1=%3d b2=%3d\n",
           scan,scan->time,scan->status,scan->byte1,scan->byte2);
 }
}
*/
	if (replace) {
		if (list) {
			for ( time=onoff ; time ; ) {
				while (!time->byte2) {
					time = time->next;
					if (!time) break;
				}
				if (time && time->next) {
					begin = time->time;
					time = time->next;
					while (time->byte2) {
						time = time->next;
						if (!time) break;
					}
					if (time) {
						end = time->time;
//dprintf("splice() begin=%ld end=%ld\n",begin,end);
						for ( scan=clip->events.first ; scan ; scan=scan->next ) {
							if (scan->status == status) {
								if ((status==MIDI_CCHANGE) && (scan->byte1 != cc)) {
									continue;
								}
								if ((scan->time >= begin) && (scan->time < end)) {
//dprintf("scan found event=%8X time=%8ld status=%2X b1=%3d b2=%3d\n",
//         scan,scan->time,scan->status,scan->byte1,scan->byte2);
									if (replace) scan->time = -1;
								}
							}
						}
					}
				}
				else break;
			}
		}
	}
	list = (struct Event *)functions->List_Cat(
		             (struct ListItem *)clip->events.first,(struct ListItem *)list);
	list = functions->sorteventlist(list);
/*
if (list) {
 dprintf("splice list:\n");
 for (scan=list ; scan ; scan=scan->next) {
  dprintf("  event=%8X time=%8ld status=%2X b1=%3d b2=%3d\n",
           scan,scan->time,scan->status,scan->byte1,scan->byte2);
 }
}
*/
	for ( scan=list ; scan ; scan=list ) {
		if (scan->time >= 0) break;
		else {
			list = scan->next;
			functions->freeevent(scan);
		}
	}
	clip->events.first = list;
	clip->events.point = list;
	functions->freelist(onoff);
}

/*==========================================================================
* splicecontrol
*
*  INPUT: pointers to track and tool
* RETURN: none
*
* EFFECT: inserts control messages in track
*=========================================================================*/

static void __regargs splicecontrol(struct Track *track, struct JoyTool *tool)

{
	struct Clip *clip;
	UBYTE replace;
	char *toolname;
	char prompt[55] = {
		"Save the latest meanderings from "
	};

	if (track->edit && track->edit->inedit) clip = &track->edit->clip;
	else clip = &track->clip;
	Forbid();
	if (clip->locked) {
		Permit();
#pragma msg 104 ignore push
		functions->msgbox("StickIt cannot save; clip locked.","Poo",NULL,NULL,1,
	                     BP_LGREY,BP_RED,BP_BLUE);
#pragma msg 104 ignore pop
//dprintf("clip locked! freeing lists\n");
		if (tool->xonoff) functions->freelist(tool->xonoff);
		if (tool->yonoff) functions->freelist(tool->yonoff);
		if (tool->xlist) functions->freelist(tool->xlist);
		if (tool->ylist) functions->freelist(tool->ylist);
		return;
	}
//else dprintf("splicing: xlist=%8X ylist=%8X\n",tool->xlist,tool->ylist);
	clip->locked = 1;
	Permit();
	if (!(tool->flags & AUTOSAVE)) {
		toolname = MakeEditWindowTitle(&tool->tool,WTITLE,20);
		sprintf(&prompt[33],"%s?",(toolname)?toolname:"StickIt");
#pragma msg 104 ignore push
		replace = (UBYTE)functions->msgbox(prompt,
		                        "Mix","Replace","Nope",1,BP_LGREY,BP_BLACK,BP_BLUE);
#pragma msg 104 ignore pop
		functions->myfree(toolname,strlen(toolname)+1);
		if (replace == 2) {
			clip->locked = 0;
			if (tool->xonoff) functions->freelist(tool->xonoff);
			if (tool->yonoff) functions->freelist(tool->yonoff);
			if (tool->xlist) functions->freelist(tool->xlist);
			if (tool->ylist) functions->freelist(tool->ylist);
			return;
		}
	}
	else replace = 1;
	if (tool->xlist)
		splicelist(clip,tool->xlist,tool->xonoff,status[tool->xtype],tool->xcc,replace);
	if (tool->ylist)
		splicelist(clip,tool->ylist,tool->yonoff,status[tool->ytype],tool->ycc,replace);
	clip->locked = 0;
	if ((tool->xlist || tool->ylist) && track->edit->window) // refresh edit window
		functions->updateobject(&functions->inittempo,0);

	return;
}

/**************************************************************************/
/** looping transport handler **/
/*******************************/
static void __inline dotransport(struct Track *track, struct JoyTool *tool,
                                 long command, long time, long end, long offset)
{
	if (!tool->tool.touched) inittool(tool);
	switch (command) {
		case TC_STOP :
			tool->flags &= ~RUNNING;
			if (tool->xlist || tool->ylist) splicecontrol(track,tool);
			tool->xlist = tool->ylist = NULL;
			tool->xonoff = tool->yonoff = NULL;
			break;

		case TC_START :
			tool->oldx = tool->oldy = -1;
			tool->newx = tool->newy = -1;
		case TC_PLAY  :
//			tool->xlist = tool->ylist = NULL;
//			tool->xonoff = tool->yonoff = NULL;
			tool->flags |= RUNNING;
			break;

		case TC_TICK :
			break;
	}
}

/*==========================================================================
* transportcode
*
*  INPUT: command, time, endtime, offset
* RETURN: none
*
* EFFECT: looping transport skeleton
*=========================================================================*/

static void __saveds __stdargs transportcode(command, time, end, offset)

register long command;
register long time;
register long end;
long offset;

{
	register struct JoyTool *tool;
	register struct Track *track;

	if ((time -= offset) < 0) time = 0;
/*	if (functions->multiin) */
	track = (struct Track *)functions->tracklist;
/*	else track = master.intrack;*/
	for ( ; track ; track=track->next ) {
		tool = (struct JoyTool *)track->toollist;
		for ( ; tool ; tool=(struct JoyTool *)tool->tool.next ) {
			if (tool->tool.toolid == Stkt_TOOLID) {
				dotransport(track,tool,command,time,end-offset,offset);
			}
		}
	}
}

/*==========================================================================
* removetool
*
*  INPUT: none
* RETURN: none
*
* EFFECT: removetool() is called when B&P removes the ToolMaster from the
*         system.  called when B&P is closing down, or RemoveTool is selected
*         from the Toolbox.  Usually not needed.
*=========================================================================*/

static void __saveds __stdargs removetool(void)

{
	functions->removetransport(transportcode);
}

/*==========================================================================
* inittoolmaster
*
*  INPUT: none
* RETURN: pointer to ToolMaster
*
* EFFECT: initializes.
*=========================================================================*/

struct ToolMaster * __saveds __stdargs inittoolmaster(void)

{
	memset((char *)&master,0,sizeof(struct ToolMaster));
	master.toolsize = TOOL_SIZE;
	master.toolid = Stkt_TOOLID;
	master.edittool = edittool;
	master.processevent = processevent;
//	master.processlist = processlist;
	master.removetool = removetool;
	master.loadtool = loadtool;
	master.image = &JoyIcon;
	strcpy(master.name,WTITLE);
	master.tooltype = TOOL_NORMAL;
	memset((char *)&ontimemaster,0,sizeof(struct ToolMaster));
	ontimemaster.processevent = processontime;
	ontimemaster.tooltype = TOOL_NORMAL | TOOL_ONTIME;
	functions->installtransportpl((void *)transportcode,0);
	return(&master);
}
